import numpy
import math
import datetime
from matplotlib import pyplot
import ohtools
from scipy import signal

NA_LIM = 0.5

def get_ohstats(filepath, printflag):
	
	try:
		numpy.loadtxt(filepath, dtype=str)
	except IOError:
		return 0

	#Loading data:
	times = numpy.loadtxt(filepath, dtype=str, usecols=(0,))
	temps31 = numpy.loadtxt(filepath, usecols=(1,))
	temperr31 = numpy.loadtxt(filepath, usecols=(2,))
	intens31 = numpy.loadtxt(filepath, usecols=(3,))
	intenserr31 = numpy.loadtxt(filepath, usecols=(4,))

	temps42 = numpy.loadtxt(filepath, usecols=(5,))
	temperr42 = numpy.loadtxt(filepath, usecols=(6,))
	intens42 = numpy.loadtxt(filepath, usecols=(7,))
	intenserr42 = numpy.loadtxt(filepath, usecols=(8,))

	q31 = numpy.loadtxt(filepath, usecols=(9,))
	q42 = numpy.loadtxt(filepath, usecols=(10,))
	na = numpy.loadtxt(filepath, usecols=(11,))

	numofmeas = times.size

	tmfmt = '%Y-%m-%dT%H:%M:%S' #Time format:
	#Transforming to dates:
	timevec = numpy.zeros(numofmeas, dtype=datetime.datetime)
	for i in range(numofmeas):
		timevec[i] = datetime.datetime.strptime(times[i][2:21], tmfmt)

	start_date = times[0][2:12]
	end_date = times[-1][2:12]

	lim = 25
	indx = numpy.arange(lim, numofmeas-lim)
	nofres = indx.size

	#Temperature weighted averages
	mt31, stdvmean31 = wgt_mean(temps31[indx], temperr31[indx])
	mt42, stdvmean42 = wgt_mean(temps42[indx], temperr42[indx])

	#Average relative errors
	avgterr31, avgrelterr31 = meanerror(temps31[indx], temperr31[indx])
	avgterr42, avgrelterr42 = meanerror(temps42[indx], temperr42[indx])

	#Intensity weighted averages
	mi31, stdvi31 = wgt_mean(intens31[indx], intenserr31[indx])
	mi42, stdvi42 = wgt_mean(intens42[indx], intenserr42[indx])	

	#Wave parameters:
	t_std31 = std(temps31[indx])
	t_std42 = std(temps42[indx])
	i_std31 = std(intens31[indx])
	i_std42 = std(intens42[indx])

	#Average Na intensity
	avgna = ohtools.mean(na[indx])
	
	wptemp31 = ohtools.wavepower(timevec[indx], temps31[indx])
	wptemp42 = ohtools.wavepower(timevec[indx], temps42[indx])

	wpintens31 = ohtools.wavepower(timevec[indx], intens31[indx])
	wpintens42 = ohtools.wavepower(timevec[indx], intens42[indx])

	stats = [mt31, stdvmean31, avgterr31, mi31, stdvi31, t_std31, i_std31, wptemp31, wpintens31,\
			 mt42, stdvmean42, avgterr42, mi42, stdvi42, t_std42, i_std42, wptemp42, wpintens42, avgna]

	if(printflag == 1):
		print('--------------------------------------------------------------------------')
		print('OH statistics for {} / {}'.format(start_date, end_date))
		print('--------------------------------------------------------------------------')
		print('Band:                      (3,1)              (4,2)         unit')         
		print('--------------------------------------------------------------------------')
		print('Avg T: {:>25.2f} {:>19.2f}          K'.format(mt31, mt42))
		print('Std. of mean: {:18.2f} {:19.2f}          K'.format(stdvmean31, stdvmean42))
		print('Avg. error: {:>12.2f} / {:>5.2f}       {:>5.2f} / {:>5.2f}          K / %'.format( \
      	avgterr31, avgrelterr31, avgterr42, avgrelterr42))
		print('Avg I: {:>25.2f} {:>19.2f}          counts/sec'.format(mi31, mi42))
		print('Std. of mean: {:>18.2f} {:19.2f}          counts/sec'.format(stdvi31, stdvi42))
		print('Std. of T: {:>21.2f} {:>19.2f}          K^2'.format(t_std31, t_std42))
		print('Std. if I: {:>21.2f} {:>19.2f}          (counts/sec)^2'.format(i_std31, i_std42))
		print('T wavepower: {:>19.2f} {:>19.2f}          [a.u]'.format(wptemp31, wptemp42))
		print('I wavepower: {:>19.2f} {:>19.2f}          [a.u]'.format(wpintens31, wpintens42))
		print('-------------------------------------------------------------------------:')      
		print('Avg. Na-level: {:>6.2f} counts/sec'.format(avgna))
		print('--------------------------------------------------------------------------')
		return 1
	else:
		return stats	


def wgt_mean(values, errors):
	#WGT_MEAN: Calculates the weighted mean based on temperatures within certain limits. Errorlim in percent, Nalim in counts/sec

	nofvals = len(values)
	sumofw = 0
	wgmean = 0

	for i in range(nofvals):
		if math.isnan(values[i]): pass
		else:	
			w = 1/errors[i]**2
			sumofw += w
			wgmean += w*values[i]

	if(sumofw != 0 and wgmean != 0):
		return wgmean/sumofw, numpy.sqrt(1/sumofw)
	else:
		return float('nan'), float('nan')

def meanerror(values, errors):
	
	meanerr = 0
	meanrelerr = 0
	nofvals = values.size

	notnan  = 0
	for i in range(nofvals):
		if math.isnan(values[i]): pass
		else:
			meanerr += errors[i]
			meanrelerr += errors[i]/values[i]
			notnan += 1

	if(meanerr != 0 and meanrelerr != 0):
		return meanerr/notnan, meanrelerr/notnan*100
	else:
		return float('nan'), float('nan')


def mean(data):
	#OHMEAN: Returns the average temperature in a tempseries. Disregards the nan 
	#temperatures. 
	valid = []
	for i in data:
		if(math.isnan(i) == False):
			valid.append(i)
	
	return numpy.mean(valid)

def std(data):
	#STD: Returns the standard deviation of a data series. Disregards the nan data.
	
	valid = []
	for i in data:
		if(math.isnan(i) == False):
			valid.append(i)
	
	return numpy.std(valid)

	
def var(data):
	#VAR: Returns the variance of data vector. Disregards the nan data.
	
	valid = []
	for i in data:
		if(math.isnan(i) == False):
			valid.append(i)
	
	return numpy.var(valid)

def smooth_wgt(values, errors, span=5):

	nofvals = len(values)
	smoothed = numpy.zeros(nofvals)
	smootherr = numpy.zeros(nofvals)
	
	sphalf = span/2
	end_lim = math.floor(sphalf)
	span_u = math.ceil(sphalf)
	span_l = math.floor(sphalf)

	weights = numpy.zeros(nofvals)
	for i in range(nofvals):
		weights[i] = 1/errors[i]**2

	k = 0
	l = span_l	
	for i in range(nofvals):
		if i < end_lim:
			interval = numpy.arange(i-k,i+span_u)
			weightsum = numpy.sum(weights[interval])
			smoothed[i] = numpy.sum(weights[interval]*values[interval])/weightsum
			smootherr[i] = numpy.sqrt(weightsum)
			k += 1
		elif i > nofvals-1-end_lim:
			interval = numpy.arange(i-span_l,i+l)
			weightsum = numpy.sum(weights[interval])
			smoothed[i] = numpy.sum(weights[interval]*values[interval])/weightsum
			smootherr[i] = numpy.sqrt(weightsum)
			l -= 1
		else:
			interval = numpy.arange(i-span_l,i+span_u)
			weightsum = numpy.sum(weights[interval])
			smoothed[i] = numpy.sum(weights[interval]*values[interval])/weightsum
			smootherr[i] = numpy.sqrt(weightsum)

	return smoothed, smootherr


def wavepower(timevec, data):

	nofmes = data.size
	nofnan = count_nan(data)

	if nofnan > 0.33*nofmes: return float('nan')

	pgram, periods, dvar = lombscargle_periodogram(timevec, data) 

	#Wavepower limits
	ll = 0
	for i, item in enumerate(periods):
		if item >= 5:
			ll = i
			break
	ul = 0 
	for i, item in enumerate(periods):
		if item >= 0.8*nofmes:
			ul = i
			break

	noiselev = numpy.mean(pgram[0:ll])
	wp = numpy.sum(pgram[ll:ul]-noiselev)*dvar

	if wp < 0: wp = float('nan')

	return wp

def lombscargle_periodogram(timevec, data):

	x, y = assign_minutes(timevec, data)
	N = x.size

	y -= numpy.mean(y)
	y *= numpy.hamming(N)

	yvar = numpy.var(y)

	ofac = 4
	hifac = 1.5
	nout = (ofac*hifac)/2*N 

	minfreq = 1/(x[-1]-x[0])
	maxfreq = hifac*N/(2*(x[-1]-x[0]))

	freq = numpy.linspace(minfreq, maxfreq, nout)
	w = 2*numpy.pi*freq

	periods = 1/freq[::-1]

	pdgnonnorm = signal.lombscargle(x, y, w)

	pdg = numpy.sqrt(4*(pdgnonnorm[::-1]/N))

	return pdg, periods, yvar

def assign_minutes(timevec, data):
	
	minutes = [] #Number of minutes since start
	newdata = [] #New datavector
	m = 0 #Minute count

	for i, time in enumerate(timevec):
		
		if i < len(timevec)-1: 
			td = timevec[i+1]-time

		if td.seconds >= 120:
			m += 1

		if math.isnan(data[i]): 
			pass
		else: 
			minutes.append(m)
			newdata.append(data[i])

		m += 1

	return numpy.array(minutes, dtype=numpy.float64), numpy.array(newdata, dtype=numpy.float64)

def count_nan(data):

	nanc = 0
	for i in data:
		if math.isnan(i): nanc += 1

	return nanc


def write_daystats(f, date, stats):
	
	f.write(date.isoformat() + 'Z')
	for i in stats:
		f.write('{:>15.2f}'.format(i))
	f.write('\n')

def dmnum2str(dmnum):
	if(dmnum < 10):
		return '0' + str(dmnum)
	else:
		return str(dmnum)	


def set_filename(date):
	
	nextdate = date + datetime.timedelta(1)
	nextday = nextdate.day
	 
	return 'OH' + str(date.year) + '_' + dmnum2str(date.month) + '_' + dmnum2str(date.day) + dmnum2str(nextday)

def print_scan_result(sign, totsign, t31, et31, i31, ei31, it31, t42, et42, i42, ei42, it42, na, naohratio, time, err_message_31, err_message_42):
	#PRINT_SCAN_RESULT: Prints the temperatures, errors, time and progress
	
	scans = 'Scan {:>3d} of {:>3d} '.format(sign, totsign)
	result_31 = '(3,1): T: {0:>5.1f} ± {1:>4.1f} I: {2:>7.2f} ± {3:>5.1f} it: {4:>2d}'.\
				format(t31, et31, i31, ei31, it31)
	result_42 = '(4.2): T: {0:>5.1f} ± {1:>4.1f} I: {2:>7.2f} ± {3:>5.1f} it: {4:>2d}'.\
				format(t42, et42, i42, ei42, it42)
	nastr = 'Na: {0:>7.2f} {1:>7.2f}'.format(na, naohratio)
	timestr = time.isoformat()
	
	err_string = ''
	if(len(err_message_31) > 0):
		err_string += 'Err (3,1): {}'.format(err_message_31)
		if(len(err_message_42) > 0):
			err_string += '  '
	if(len(err_message_42) > 0):
		err_string += 'Err (4,2): {}'.format(err_message_42)
	
	print(scans + result_31 + '  ' + result_42 + '  ' + nastr + '  ' + timestr + '  ' + err_string)

def write_scan_result(t31, et31, i31, ei31, t42, et42, i42, ei42, q31, q42, na, time, file):
	#WRITE_SCAN_RESULT: Writes results from scan to file
	
	file.write(time.isoformat() + 'Z ')
	file.write('  ')
	file.write('{:>6.2f} {:>6.2f} {:>7.2f} {:>7.2f}'.format(t31, et31, i31, ei31)) #Writing (3,1) results
	file.write('  ')
	file.write('{:>6.2f} {:>6.2f} {:>7.2f} {:>7.2f}'.format(t42, et42, i42, ei42)) #Writing (4,2) results
	file.write('  ')
	file.write('{:>7.2f} {:>7.2f} {:>7.2f}'.format(q31, q42, na))
	file.write('\n')