import numpy
from matplotlib import pyplot
from scipy import signal

def avgspec(data, start, stop, filtflag=0):
	#Function avgsig: returns the averaged signal between start and stop indices
	avgspectrum = numpy.zeros(1024)
	sigindices = numpy.arange(start,stop)
	nofsig = len(sigindices)
	
	if(filtflag == 1):
		spectrums = numpy.zeros((nofsig, 1024))
		for i in sigindices:
			spectrums[i-start] = signal.medfilt(data[i,11:1035], 3)
		
		for i in range(1024):
			avgsig[i] = sum(spectrums[:,i])/nofsig
	else:
		for i in range(1024):
			avgspectrum[i] = sum(data[start:stop,11+i])/nofsig
	
	return avgspectrum


def avgspec_nafilt(signal, start, stop, nalimit):
	
	avgsig = numpy.zeros(1024)

	nofclf = 0
	for i in numpy.arange(start, stop):
		if naoh_ratio(signal[i,11:1035]) < nalimit:
			nofclf += 1

	clf = numpy.zeros((nofclf, 1024))

	k = 0
	for i in numpy.arange(start, stop):
		spec = signal[i,11:1035]
		if naoh_ratio(spec) < nalimit:
			clf[k,:] = spec
			k += 1  

	avgspectrum = numpy.zeros(1024)

	for i in range(1024):
		avgspectrum[i] = numpy.sum(clf[:,i])/nofclf

	return avgspectrum

def naoh_ratio(spectrum):

	na = sodiumlev(spectrum)
	q42 = qbranch(spectrum)[1]

	try:
		na/q42
	except ZeroDivisionError:
		return float('nan')

	return na/q42


def band_borders(peaks):
	#BAND_BORDERS: Returns the borders of the (3,1) and (4,2) band in the following
	#order: left 31, right 31, left 42, right 42

	return [peaks[0]-14, peaks[5]+20, peaks[6]-15, peaks[11]+20]

def basepts(peaks):
	#BASEPTS: Returns the defined no signal points.

	#(3,1) Basepts:
	leftp22 = numpy.arange(peaks[0]-15, peaks[0]-10)
	rightp22 = numpy.arange(peaks[0]+15, peaks[0]+18)

	rightp13 = numpy.arange(peaks[3]+15, peaks[3]+22)
	rightp14 = numpy.arange(peaks[5]+14, peaks[5]+18)

	basepts31 = numpy.r_[leftp22,rightp22,rightp13,rightp14]

	#(4,2) Basepts:
	leftp22 = numpy.arange(peaks[6]-15, peaks[6]-10)
	rightp22 = numpy.arange(peaks[6]+15, peaks[6]+21)

	rightp13 = numpy.arange(peaks[9]+25, peaks[9]+28)
	rightp14 = numpy.arange(peaks[11]+16, peaks[11]+22)

	basepts42 = numpy.r_[leftp22, rightp22, rightp13, rightp14]

	return basepts31, basepts42

def basecurve(spectrum, basepts, left_pix, right_pix, toggleplot=0):
	#BASECURVE: Fits a 3rd degree polynimial through the given basepoints

	c = numpy.polyfit(basepts, spectrum[basepts], 3)
	baseshape = numpy.zeros(len(spectrum))
	
	for i in numpy.arange(left_pix, right_pix):
		baseshape[i] = c[0]*i**3 + c[1]*i**2 + c[2]*i + c[3]
	
	if(toggleplot == 1):
		pltpts = numpy.arange(left_pix, right_pix)
		pyplot.figure(figsize=(4*1.618, 4))
		pyplot.plot(pltpts, spectrum[pltpts])
		pyplot.plot(pltpts, baseshape[pltpts], 'r')
		pyplot.plot(basepts, spectrum[basepts], 'g+')
		pyplot.ylim((-2,20))
		pyplot.xlabel('Pixel number')
		pyplot.ylabel('Intensity [counts/sec]')
		pyplot.show()
	
	return baseshape[left_pix:right_pix]


def locate_peaks(spectrum, toggleplot=0):
	#LOCATE_PEAKS: Returns the peaks from the average spectrum
	#Input: Spectrum, array of size 1024
	#Output: Array with peak positions

	#Subtracing a constant offset
	offsetpts = numpy.r_[numpy.arange(194,203), numpy.arange(320,322), numpy.arange(730,734)]
	offset = numpy.mean(spectrum[offsetpts])
	spectrum -= offset

	lines = numpy.array([305, 335, 357, 380, 412, 432, 717, 747, 773, 799, 837, 859], dtype=int) #Predefined line positions
	instrf = numpy.loadtxt('ohtools/instrfunction.dat')
	n = len(instrf)
	
	peaks = numpy.zeros(len(lines), dtype=int)

	for index, linepos in enumerate(lines):

		shift = 3 #Number of pixels away from peaks to start
		start = linepos-shift-14

		scale = numpy.max(spectrum[linepos-4:linepos+5])
		scaled_instr = scale*instrf/numpy.max(instrf)

		x = numpy.arange(4,n-11)
		r = spectrum[start+x] - scaled_instr[x]

		sse = 0
		newsse = 0

		for i in r: sse += i**2 

		while True:

			k = start+1
			r = spectrum[k+x] - scaled_instr[x] #Residuals
			newsse = 0

			for i in r: 
				newsse += i**2

			if newsse < sse: #If sse has decreased, proceed. Else stop.
				start += 1
				sse = newsse
			else:
				break

		peaks[index] = start + 14 #Position of the peak

	if toggleplot == 1:	
		pyplot.plot(numpy.arange(0,1024), spectrum)

		for i in peaks:
			scale = max(spectrum[i-3:i+4])
			line = scale*instrf/max(instrf)
			pyplot.plot(numpy.arange(i-14,i-14+n), line, 'r')
			rangepts = numpy.array([i-(14-x[0]), i+(x[-1]-14)])
			pyplot.plot(rangepts, spectrum[rangepts], 'k.')

		pyplot.show()

	return peaks

def filtsig(data):
	#FILTSIG: Filters all the spectrums in the signal matrix
	
	nsig = data.shape[0]
	fs = numpy.zeros((nsig,1035))
	for i in range(nsig):
		fs[i,0:10] = data[i,0:10]
		fs[i,11:1035] = signal.medfilt(data[i,11:1035], 3)

	return fs

def qbranch(spectrum):
	#QBRANCH: Returns integrated intensities of the (3,1) and (4,2) Q-branches
		
	q31pts = numpy.arange(231,274) #Integration points
	q42pts = numpy.arange(631,668)
	bspts31 = numpy.arange(196,230)
	bspts42 = numpy.arange(591,630)
	
	base31 = numpy.mean(spectrum[bspts31])
	base42 = numpy.mean(spectrum[bspts42])
	 
	q31int = numpy.sum(spectrum[q31pts]-base31)
	q42int = numpy.sum(spectrum[q42pts]-base42)

	if(q31int < 0): q31int = 0
	if(q42int < 0): q42int = 0
	
	return q31int, q42int	
		
def sodiumlev(spectrum):
	#SODIUMLEV: Returns the integrated Na intensity
	
	sodiumpts = numpy.arange(910,1000)
	basepts1 = numpy.arange(876,900)
	basepts2 = numpy.arange(1001,1024)
	basepts = numpy.concatenate((basepts1, basepts2))
	base = numpy.mean(spectrum[basepts]) 
	
	na = numpy.sum(spectrum[sodiumpts]-base) #Not allowed to be negative for convenince
	if(na < 0):
		na = 0

	return na
