import numpy
from matplotlib import pyplot
import math
import time

lineinfo = numpy.loadtxt("ohtools/lineinfo.txt")
wlen = lineinfo[:,0] #Wavelengths
init_e = lineinfo[:,1] #Initial energies
lstr = lineinfo[:,2] #Line strengths
freq = lineinfo[:,3] #Frequencies
trc_winter = lineinfo[:,4]
trc_summer = lineinfo[:,5]

response_kr = numpy.loadtxt('ohtools/response_kr.dat')

constants = numpy.loadtxt('ohtools/constants')
ekb = constants[:,0] #e*hckb
fr3s = constants[:,1] #fr**3*s

#Energies and linestrengts to calculate partition function
ekb31 = numpy.loadtxt('ohtools/ekb31')
ekb42 = numpy.loadtxt('ohtools/ekb42')
lstr31 = numpy.loadtxt('ohtools/linestr31.txt')
lstr42 = numpy.loadtxt('ohtools/linestr42.txt')

#Error limits
rel_temperr_lim = 5 # %
temperr_lim = 10 # K
intenserr_lim = 50 #Counts/sec
rel_intenserr_lim = 10 # %

#Sensible limits on temperature and intensity
templim_low = 150
templim_upper = 270
intenslim_low = 100
intenslim_upper = 1500

ACCLIM = 1e-3
ITMAX = 10

def instrfunc(peaks, borders):
	#INSTRFUNC: Return a matrix with the instrument function for each line
	
	lineshape = numpy.loadtxt("ohtools/instrfunction.dat")
	lineshppts = len(lineshape)
	
	w_31 = numpy.zeros((borders[1]-borders[0], 6))
	w_42 = numpy.zeros((borders[3]-borders[2], 6))
	
	for i in range(6):
		k = peaks[i]-borders[0]-14 #Start index for the line
		for j in range(lineshppts):
			w_31[k+j,i] = lineshape[j]
		
	for i in numpy.arange(6):
		k = peaks[i+6]-borders[2]-14
		for j in range(lineshppts):
			w_42[k+j,i] = lineshape[j]
	
	return w_31, w_42

def instrfunc_kr(peaks, borders):
	#INSTRFUNC_KR: Computes the instrument functions for each line multiplied with the
	#response curve, linestrengths, frequencies and transmisson coefficients.
	
	w_31, w_42 = instrfunc(peaks, borders)
	
	for i in range(6):
		w_31[:,i] *= response_kr[borders[0]:borders[1]]*fr3s[i]/trc_winter[i]
	
	for i in range(6):
		w_42[:,i] *= response_kr[borders[2]:borders[3]]*fr3s[i+6]/trc_winter[i+6]
	
	return w_31, w_42	

def expwmat(w, tau, shift):
	
	n, m = w.shape
	wmod = numpy.zeros((n,m))
	for i in range(6):
		wmod[:,i] = w[:,i]*numpy.exp(-tau*ekb[i+shift])
	
	return wmod

def ohmodel(rho, wmod):
	
	modlen = wmod.shape[0]
	model = numpy.zeros(modlen)
	for i in range(modlen):
		model[i] = numpy.sum(wmod[i,:])
	
	model *= rho
	
	return model
	
def jacobitau(wmod, tau, qr, shift):
	
	if(shift == 0):
		dq = delpartfunc_31(tau)
	else:
		dq = delpartfunc_42(tau)
		
	n = wmod.shape[0]
	jtm = numpy.zeros((n,6))
	jt = numpy.zeros(n)
	
	for i in range(6):
		jtm[:,i] = -(ekb[i+shift]+dq/qr)*wmod[:,i]
	
	for i in range(n):
		jt[i] = numpy.sum(jtm[i,:])
	
	return jt
			
def partfunc_31(tau):
	#PARTFUNC: Returns the partition function for a given tau
	
	qr = 0

	for i in numpy.arange(0,20,2):
		qr += lstr31[i]*numpy.exp(-ekb31[i]*tau)

	for i in numpy.arange(1,20,2):
		qr += lstr31[i]*numpy.exp(-ekb31[i]*tau)
	
	return qr


def delpartfunc_31(tau):
	
	delqr = 0

	for i in numpy.arange(0,20,2):
		delqr += -ekb31[i]*lstr31[i]*numpy.exp(-ekb31[i]*tau)

	for i in numpy.arange(1,20,2):
		delqr += -ekb31[i]*lstr31[i]*numpy.exp(-ekb31[i]*tau)
	
	return delqr

def partfunc_42(tau):
	#PARTFUNC_42: Returns the partition function for a given tau for the (4,2) band
	
	qr = 0

	for i in numpy.arange(0,20,2):
		qr += lstr42[i]*numpy.exp(-ekb42[i]*tau)

	for i in numpy.arange(1,20,2):
		qr += lstr42[i]*numpy.exp(-ekb42[i]*tau)
	
	return qr	

def delpartfunc_42(tau):
	
	delqr = 0
	for i in numpy.arange(0,20,2):
		delqr += -ekb42[i]*lstr42[i]*numpy.exp(-ekb42[i]*tau)

	for i in numpy.arange(1,20,2):
		delqr += -ekb42[i]*lstr42[i]*numpy.exp(-ekb42[i]*tau)
	
	return delqr
	
def get_temp_fitbg(spectrum, tguess, w, bandflag, infoflag, baseshape):
	
	err_message_gn = '' #Error message from Gauss-Newton iteration
	err_message = ''
	
	speclen = spectrum.size
	alphaguess = 1.0
	betaguess = 1.0
	uncrt = numpy.zeros((5,2))
	
	#Deciding model function from the bandflag
	if(bandflag == 0):
		shift = 0
		plot_title = 'OH (3,1) band'
		partfunc = partfunc_31
		delpartfunc = delpartfunc_31
	elif(bandflag == 1):
		shift = 6
		plot_title = 'OH (4,2) band'
		partfunc = partfunc_42
		delpartfunc = delpartfunc_42

	start = time.clock()	
	tauguess = 1/tguess
	qr = partfunc(tauguess)
	wmod = expwmat(w, tauguess, shift)/qr #Divide by partition function
	specshape = ohmodel(1, wmod)
	
	#Scaling the band to match the measured spectrum:	
	rho_initial = max(spectrum - alphaguess*baseshape - betaguess)/max(specshape)
 
	ohband = rho_initial*specshape
	model = ohband + alphaguess*baseshape + betaguess
	intens_initial = numpy.sum(spectrum - alphaguess*baseshape - betaguess)
	
	residuals = numpy.zeros(speclen)
	for i in range(speclen):
		residuals[i] = spectrum[i] - model[i]
	
	sse_initial = 0 #Initial sum of square erros
	for i in residuals:
		sse_initial += i**2
 
	ds = 1e6
	itcount = 0
	
	sse = sse_initial
	params = numpy.array([rho_initial, tauguess, alphaguess, betaguess])
	
	#Gauss-Newton algorithm
	while(ds > ACCLIM and itcount < ITMAX):
		
		jacobi = numpy.zeros((speclen,4)) #Jacobi matrix
		jacobi[:,0] = specshape
		jacobi[:,1] = params[0]*jacobitau(wmod, params[1], qr, shift)
		jacobi[:,2] = baseshape
		jacobi[:,3] = 1
		
		#jacobi[:,1] *= params[0] #Multiplying with rho
		
		jtj = jacobi.T.dot(jacobi) #Calculating (J^T*J)
		try:
			#Calculating (J^T*J)^-1*J^Tr:
			invjtjjtr = numpy.linalg.solve(jtj,jacobi.T.dot(residuals))
		except numpy.linalg.LinAlgError:
			err_message_gn = 'Normal eqn. failure'
			break
	
		params[0] += invjtjjtr[0] #Adding changes
		params[1] += invjtjjtr[1] 
		params[2] += invjtjjtr[2]
		params[3] += invjtjjtr[3]
		
		qr = partfunc(params[1])
		wmod = expwmat(w, params[1], shift)/qr #New model with the adjusted parameters
		specshape = ohmodel(1, wmod)
		ohband = params[0]*specshape
		model = ohband + params[2]*baseshape + params[3]
		
		oldsse = sse
		residuals = numpy.zeros(speclen)
		for i in range(speclen):
			residuals[i] = spectrum[i] - model[i]
		
		sse = 0
		for i in residuals:
			sse += i**2	
			
		ds = abs(oldsse - sse) #Change in sum of square errors
		itcount += 1
	
	elapsed = time.clock()-start
	if(len(err_message_gn) == 0): 
	#If err_message_gn != 0 Gauss-Newton algortim succeeded. Check values and calculate
	#uncertainties
	
		newt = 1/params[1] #New temperature
		intens_mod = numpy.sum(ohband)
		intens_real = numpy.sum(spectrum - params[2]*baseshape - params[3])
		#intens_real = numpy.sum(spectrum)
		
		#Calculating errors:
		uncrt, errmess_cov = uncertainties_fitbg(jtj, residuals, params, intens_real, baseshape, speclen)
		
		intens_mod_err = intens_mod*uncrt[1,1]/100

		#Checking result:
		if(newt < templim_low or newt > templim_upper): #Temperature limits
			newt = float('nan')
			uncrt[0,1] = float('nan')
			uncrt[0,0] = float('nan')
			intens_mod_err = float('nan')
			err_message += 'Not valid temp, '
		
		if(intens_mod < intenslim_low or intens_mod > intenslim_upper): #Intensity limits
			intens_real = float('nan')
			intens_mod = float('nan')
			intens_mod_err = float('nan')
			uncrt[2,:] = float('nan')
			uncrt[1,:] = float('nan')
			err_message += 'Not valid intensity, '
		
		#Checking errors:	
		if(uncrt[0,0] > temperr_lim or uncrt[0,1] > rel_temperr_lim): #Temperature error (set to 25 K)
			uncrt[0,:] = float('nan')
			newt = float('nan')
			err_message += 'High T error, '
		
		if(intens_mod_err > intenserr_lim or uncrt[1,1] > rel_intenserr_lim): #Intensity error
			intens_real = float('nan')
			intens_mod = float('nan')
			intens_mod_err = float('nan')
			err_message += 'High I error'
					
	else:
		newt = float('nan')
		intens_real = float('nan')
		intens_mod = float('nan')
		uncrt[:,:] = float('nan')
		intens_mod_err = float('nan')
		err_message = err_message_gn #Setting error message from Gauss-Newton calculation
		
	if(infoflag == 1):
		print('-------------------Before iteration-------------------')
		print('Initial temperature: {:4.2f}'.format(tguess))
		print('Initial intensity: {:6.2f}'.format(intens_initial))
		print('Initial rho: {:6e}'.format(rho_initial))
		print('Initial alpha: {:4.2f}'.format(alphaguess))
		print('Initial beta: {:2.2f}'.format(betaguess))
		print('Initial SSE: {:6.2f}'.format(sse_initial))			 	   
		print('--------------------After iteration-------------------')
		print('New temperature: {:4.2f} ± {:3.2f}, acc: {:4.2f}%'.format(newt, \
		      uncrt[0,0], uncrt[0,1]))
		print('Real intensity:  {:5.2f} ± {:3.2f}, acc: {:2.2f}%'.format(intens_real, \
			  uncrt[4,0], uncrt[4,1]))
		print('Model intensity: {:5.2f} ± {:.2f}, acc: {:2.2f}%'.format(intens_mod, \
			  intens_mod*uncrt[1,1]/100, uncrt[1,1]))
		print('New rho:         {:.3e} ± {:.3e}, acc: {:2.2f} %'.format(params[0], uncrt[1,0],  uncrt[1,1]))
		print('New alpha:       {:4.2f} ± {:2.2f} acc: {:2.2f}%'.format(params[2], \
			  uncrt[2,0], uncrt[2,1]))
		print('New beta:        {:4.2f} ± {:4.2f} acc: {:4.2f}%'.format(params[3], uncrt[3,0], uncrt[3,1]))
		print('SSE: {:6.2f}'.format(sse))
		print('Number of iterations: {:2d}'.format(itcount))
		print('CPU time: {:3.2f}'.format(elapsed))
		print('Error: {}'.format(err_message))
		
		fig, ax = pyplot.subplots(1)
		ax.plot(spectrum,'b', label = 'Real')
		ax.plot(model,'r', label = 'Model')
		ax.plot(params[2]*baseshape + params[3], 'k', linewidth = 2)
		pyplot.title(plot_title)
		pyplot.legend()
		boxtext = 'Temperature: {:3.1f} ± {:3.1f} K'.format(newt, uncrt[0,0])
		boxtext += 'Intensity: {:4.1f} ± {:4.1f} counts/sec'.format(intens_real, uncrt[4,0])
		props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
		ax.text(0.03, 0.97, boxtext, transform=ax.transAxes, verticalalignment='top', bbox=props)

		fig2 = pyplot.figure()
		pyplot.plot(spectrum-model, 'k')
		pyplot.xlabel('Pixel number')
		pyplot.ylabel('Data - Model [counts/sec]')
		pyplot.show()

	return newt, uncrt[0,0], intens_mod, intens_mod_err, itcount, err_message

def get_temp_subbg(spectrum, tguess, w, bandflag, infoflag):
	
	err_message_gn = '' #Error message from Gauss-Newton iteration
	err_message = ''
	
	speclen = spectrum.size
	betaguess = 1.0
	uncrt = numpy.zeros((4,2))
	
	#Deciding model function from the bandflag
	if(bandflag == 0):
		shift = 0
		plot_title = 'OH (3,1) band'
		partfunc = partfunc_31
		delpartfunc = delpartfunc_31
	elif(bandflag == 1):
		shift = 6
		plot_title = 'OH (4,2) band'
		partfunc = partfunc_42
		delpartfunc = delpartfunc_42

	start = time.clock()
	tauguess = 1/tguess
	qr = partfunc(tauguess)
	wmod = expwmat(w, tauguess, shift)/qr #Divide by partition function
	specshape = ohmodel(1, wmod)
	
	#Scaling the band to match the measured spectrum:	
	rho_initial = numpy.max(spectrum - betaguess)/numpy.max(specshape)
 
	ohband = rho_initial*specshape
	model = ohband + betaguess
	intens_initial = numpy.sum(model)
	
	residuals = numpy.zeros(speclen)
	for i in range(speclen):
		residuals[i] = spectrum[i] - model[i]
	
	sse_initial = 0 #Initial sum of square erros
	for i in residuals:
		sse_initial += i**2
 
	ds = 1e6
	itcount = 0
	
	sse = sse_initial
	params = numpy.array([rho_initial, tauguess, betaguess])
	
	#Gauss-Newton algorithm
	while(ds > ACCLIM and itcount < ITMAX):
		
		jacobi = numpy.zeros((speclen,3)) #Jacobi matrix
		jacobi[:,0] = specshape
		jacobi[:,1] = params[0]*jacobitau(wmod, params[1], qr, shift)
		jacobi[:,2] = 1
		
		#jacobi[:,1] *= params[0] #Multiplying with rho
		
		jtj = jacobi.T.dot(jacobi) #Calculating (J^T*J)

		try:
			#Calculating (J^T*J)^-1*J^Tr:
			invjtjjtr = numpy.linalg.solve(jtj,jacobi.T.dot(residuals))
		except numpy.linalg.LinAlgError:
			err_message_gn = 'Normal eqn. failure'
			break
	
		params[0] += invjtjjtr[0] #Adding changes
		params[1] += invjtjjtr[1]
		params[2] += invjtjjtr[2] 
		
		qr = partfunc(params[1])
		wmod = expwmat(w, params[1], shift)/qr #New model with the adjusted parameters
		specshape = ohmodel(1, wmod)
		ohband = params[0]*specshape
		model = ohband + params[2]
		
		oldsse = sse
		residuals = numpy.zeros(speclen)
		for i in range(speclen):
			residuals[i] = spectrum[i] - model[i]
		
		sse = 0
		for i in residuals:
			sse += i**2	
			
		ds = abs(oldsse - sse) #Change in sum of square errors
		itcount += 1
	elapsed = time.clock()-start
	if(len(err_message_gn) == 0): 
	#If err_message_gn != 0 Gauss-Newton algortim succeeded. Check values and calculate
	#uncertainties
	
		newt = 1/params[1] #New temperature
		intens_mod = numpy.sum(ohband)
		intens_real = numpy.sum(spectrum-params[2])
		
		#Calculating errors:
		uncrt, errmess_cov = uncertainties_subbg(jtj, residuals, params, intens_real, speclen)
		
		intens_mod_err = intens_mod*uncrt[1,1]/100

		#Checking result:
		if(newt < templim_low or newt > templim_upper): #Temperature limits
			newt = float('nan')
			uncrt[0,1] = float('nan')
			uncrt[0,0] = float('nan')
			intens_mod_err = float('nan')
			err_message += 'Not valid temp, '
		
		if(intens_mod < intenslim_low or intens_mod > intenslim_upper): #Intensity limits
			intens_real = float('nan')
			intens_mod = float('nan')
			intens_mod_err = float('nan')
			uncrt[2,:] = float('nan')
			uncrt[1,:] = float('nan')
			err_message += 'Not valid intensity, '
		
		#Checking errors:	
		if(uncrt[0,0] > temperr_lim or uncrt[0,1] > rel_temperr_lim): #Temperature error (set to 25 K)
			uncrt[0,:] = float('nan')
			newt = float('nan')
			err_message += 'High T error, '
		
		if(intens_mod_err > intenserr_lim or uncrt[1,1] > rel_intenserr_lim): #Intensity error
			intens_real = float('nan')
			intens_mod = float('nan')
			intens_mod_err = float('nan')
			err_message += 'High I error'
					
	else:
		newt = float('nan')
		intens_real = float('nan')
		intens_mod = float('nan')
		uncrt[:,:] = float('nan')
		intens_mod_err = float('nan')
		err_message = err_message_gn #Setting error message from Gauss-Newton calculation
		
	if(infoflag == 1):
		print('-------------------Before iteration-------------------')
		print('Initial temperature: {:4.2f}'.format(tguess))
		print('Initial intensity: {:6.2f}'.format(intens_initial))
		print('Initial rho: {:6e}'.format(rho_initial))
		print('Initial beta: {:3f}'.format(betaguess))
		print('Initial SSE: {:6.2f}'.format(sse_initial))			 	   
		print('--------------------After iteration-------------------')
		print('New temperature: {:4.2f} ± {:3.2f}, acc: {:4.2f}%'.format(newt, \
		      uncrt[0,0], uncrt[0,1]))
		print('Real intensity:  {:5.2f} ± {:3.2f}, acc: {:2.2f}%'.format(intens_real, \
			  uncrt[2,0], uncrt[2,1]))
		print('Model intensity: {:5.2f} ± {:.2f}, acc: {:2.2f}%'.format(intens_mod, \
			  intens_mod_err, uncrt[1,1]))
		print('New rho:         {:.3e} ± {:.3e}, acc: {:2.2f} %'.format(params[0], uncrt[1,0],  uncrt[1,1]))
		print('New beta: {:9.3f} ± {:6.3f} acc: {:2.2f} %'.format(params[2], uncrt[3,0], uncrt[3,1]))
		print('SSE: {:6.2f}'.format(sse))
		print('Number of iterations: {:2d}'.format(itcount))
		print('CPU time: {:2.6f}'.format(elapsed))
		print('Error: {}'.format(err_message))
		
		offset = numpy.zeros(speclen)
		offset[:] = params[2]
		
		fig1 = pyplot.figure(figsize=(4.5,3.5))
		ax = fig1.add_subplot(111)
		ax.plot(spectrum,'b', label = 'Real')
		ax.plot(model,'r', label = 'Model')
		pyplot.ylabel('Intensity [counts/sec]')
		pyplot.xlabel('Pixel number')
		pyplot.title(plot_title)
		pyplot.legend()
		boxtext = 'Temperature: {:3.1f} ± {:3.1f} K \n'.format(newt, uncrt[0,0])
		boxtext += 'Intensity: {:4.1f} ± {:4.1f} counts/sec'.format(intens_real, uncrt[2,0])
		props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
		ax.text(0.03, 0.97, boxtext, transform=ax.transAxes, verticalalignment='top', bbox=props)

		fig2 = pyplot.figure(figsize=(4.5,3.5))
		pyplot.plot(spectrum-model, 'k')
		pyplot.xlabel('Pixel number')
		pyplot.ylabel('Data - Model [counts/sec]')
		pyplot.show()

	return newt, uncrt[0,0], intens_mod, intens_mod_err, itcount, err_message

def uncertainties_fitbg(JTJ, residuals, params, intens_real, baseshape, speclen):
	
	err_message = ''
	
	uncrt = numpy.zeros((5,2)) #Initializing error matrix
	uncrt[:,:] = float('nan')
	
	try:
		covmat = numpy.linalg.inv(JTJ) #Computing covariance matrix
	except numpy.linalg.LinAlgError:
		err_message = 'Covariance fail'
		return uncrt, err_message
	
	sigma = numpy.std(residuals) #Std of residuals
	
	temp = 1/params[1]
	dtau = numpy.sqrt(covmat[1,1])*sigma #Error in tau
	dt = temp**2*dtau 
	reldt = dt/temp*100
	
	drho= numpy.sqrt(covmat[0,0])*sigma #Rho error
	reldrho = drho/params[0]*100
	
	dalpha = numpy.sqrt(covmat[2,2])*sigma #Alpha error
	reldalpha = dalpha/params[2]*100
	
	dbeta = numpy.sqrt(covmat[3,3])*sigma #Beta error
	reldbeta = dbeta/params[3]*100
	
	#Intensity error
	dintens_real = 0.583*numpy.sqrt(speclen) + dalpha*numpy.sum(baseshape) + dbeta*speclen 	 
	reldintens_real = dintens_real/intens_real*100
	
	uncmat = numpy.array(([dt, reldt], [drho, reldrho], [dalpha, reldalpha], \
					   [dbeta, reldbeta], [dintens_real, reldintens_real]))
	
	return uncmat, err_message

def uncertainties_subbg(JTJ, residuals, params, intens_real, speclen):
	
	err_message = ''
	
	uncrt = numpy.zeros((4,2)) #Initializing error matrix
	uncrt[:,:] = float('nan')
	
	try:
		covmat = numpy.linalg.inv(JTJ) #Computing covariance matrix
	except numpy.linalg.LinAlgError:
		err_message = 'Covariance fail'
		return uncrt, err_message
	
	sigma = numpy.std(residuals) #Std of residuals
	
	temp = 1/params[1]
	dtau = numpy.sqrt(covmat[1,1])*sigma #Error in tau
	dt = temp**2*dtau 
	reldt = dt/temp*100
	
	drho= numpy.sqrt(covmat[0,0])*sigma #Rho error
	reldrho = drho/params[0]*100

	dbeta = numpy.sqrt(covmat[2,2])*sigma #Beta error
	reldbeta = dbeta/params[2]*100
	
	#Intensity error
	dintens_real = 0.583*numpy.sqrt(speclen) + dbeta*speclen 	 
	reldintens_real = dintens_real/intens_real*100
	
	uncmat = numpy.array(([dt, reldt], [drho, reldrho], [dintens_real, reldintens_real], [dbeta, reldbeta]))
	
	return uncmat, err_message

def highpass_vec(data, cutoff=15):
	#HIGHPASS: Returns the highpass filtered data
	
	n = data.size
	
	filter = numpy.ones(n)
	filter[0:cutoff] = 0
	
	ft = numpy.fft.fft(data)
	filtft = filter*ft
	filtsig = numpy.fft.ifft(filtft)
	
	return numpy.real(filtsig)

def lowpass_vec(data, cutoff=15):

	n = data.size
	
	filter = numpy.zeros(n)
	filter[0:cutoff] = 1
	
	ft = numpy.fft.fft(data)
	filtft = filter*ft
	filtsig = numpy.fft.ifft(filtft)
	
	return numpy.real(filtsig)



	
	
